    package ORG.hd.d.par;

//    DHDPhotoGallery: a system for building Web-based Multimedia galleries.
//    Copyright (C) 1997--2000 Damon Hart-Davis
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    import java.io.*; // For File and its ilk.
    import java.util.*; // For Hashtable and its ilk.

    /**The parallel-task handler.
     * This class attempts to run tasks in parallel in a simple way
     * and to make best use of available CPUs.
     * <p>
     * This is meant to be in sympathy with
     * <a href="http://www.hd.org/Damon/JavaThreadsCommentToSun19980820.html">my proposal of late 1998 to Sun</a>
     * with their bug ID 4186261.
     *
     *
     *
     *
     *
     *
     * <p>
     * When runTask() or runTasks() is called, the current predicted
     * utilisation for the current thread (if any) is stacked and
     * is replaced, in effect, by those of the new tasks, to be restored
     * when the new tasks complete, successfully or unsuccessfully.
     *
     * @author Damon Hart-Davis
     * @version 1.41 00/10/03
     */
    public final class ParTasks
        {
        /**This class cannot be constructed by others. */
        private ParTasks() { }

        /**Are we in debug mode? */
        private static final boolean debug = false;

        /**Our thread group. */
        private static final ThreadGroup tg = new ThreadGroup("ParTasks");

        /**The absolute maximum utilisation a single task can have, ie many CPUs to 100%. */
        public static final int UTILISATION_ABSMAX = 0x40000000;
        /**The maximum utilisation a simple single-threaded task can have, eg largely CPU-bound. */
        public static final int UTILISATION_MAX = 0x100;
        /**The minimum believable utilisation a single task can have, largely I/O-bound. */
        public static final int UTILISATION_MIN = 0x1;
        /**The absolute minimum utilisation a single task can have, eg 0% CPU. */
        public static final int UTILISATION_ABSMIN = 0;

        /**Default maximum number of threads to use in this VM. */
        private static final int SYS_PROP_MAXTHREADS__DEFAULT = 256;

        /**The system property name for the stipulated number of CPUs to use.
         * This must be positive.
         * <p>
         * If set to 1, this explicitly excludes parallism.
         */
        public static final String SYS_PROP_PAR = "org.hd.d.par.level";

        /**Default number of CPUs assumed. */
        private static final int SYS_PROP_PAR__DEFAULT = 1;

        /**Wraps a Collection to return a repeating iterator.
         * A call to iterator() on the wrapper will result in
         * an iterator which resets whenever it reaches the
         * end of the collection, thus repeating indefinitely.
         * Each run through will provide the ordering that the
         * underlying iterator produces, with each iterator being
         * used to completion.
         * <p>
         * Use this to run some repeating parallel task, to help
         * overlap the end of one run with the start of the next.
         * This in principle could allow two threads to run a functor
         * on two copies of the underlying task at the same time,
         * but limiting parallelism to the size of the (underlying)
         * Collection can avoid this.
         * <p>
         * Note that if immediately after starting such a
         * sequence the underlying iterator's hasNext() returns
         * false, so will the wrapper's iterator, so empty
         * collections don't cause an infinite loop.
         * <p>
         * It is possible to have the repeating stop when a condition
         * is met by deriving from this class and override the
         * willRepeat() method.  If this is false when an iterator
         * would reset, the iterator will not reset.  Note that this
         * has no effect until the underlying iterator completes.
         */
        public static class RepeatingIteratorCollection implements Collection
            {
            /**The underlying Collection. */
            private final Collection col;
            /**Wrap an existing Collection to make its Iterator repeat. */
            public RepeatingIteratorCollection(Collection c)
                {
                if(c == null) { throw new NullPointerException(); }
                col = c;
                }
            /**If true (the default), the iterators produced will repeat. */
            public boolean willRepeat()
                { return(true); }
            /**Get a (repeating) iterator.
             * This iterator is not thread-safe; synchronize on the
             * wrapped collection to make it so.
             */
            public final Iterator iterator()
                {
                return(new Iterator(){
                    /**Maintained and accessed through _repeatIfNecessary(). */
                    private Iterator _underlying = col.iterator();
                    /**Gets the underlying iterator and replaces/resets if necessary. */
                    private Iterator _repeatIfNecessary()
                        {
                        // Should possibly reset if at end...
                        if(!_underlying.hasNext())
                            {
                            if(willRepeat()) // Yes, we want to repeat.
                                {
                                // Reset with a new, rewound iterator.
                                _underlying = col.iterator();
                                }
                            }
                        return(_underlying);
                        }
                    /**Check if there is a next, rewinding if necessary.
                     * Will return false if: (a) the collection is empty
                     * and the underlying iterator always immediately
                     * returns false to its hasNext(), or (b) if
                     * we reach the end of the underlying iterator and
                     * willRepeat() returns false.
                     */
                    public boolean hasNext()
                        { return(_repeatIfNecessary().hasNext()); }
                    /**Get the next, rewinding if necessary.
                     * This will fail to return a next under the same
                     * conditions that hasNext() will return false.
                     */
                    public Object next()
                        { return(_repeatIfNecessary().next()); }
                    /**Do the underlying remove.
                     * This might be a little unwise for this sort of
                     * op, but might make sense...
                     * <p>
                     * (We must be careful to get at the existing
                     * underlying iterator and not get a new one
                     * to apply this op.)
                     */
                    public void remove()
                        { _underlying.remove(); }
                    });
                }

            // Delegate other methods of Collection.
            public boolean add(Object o) { return(col.add(o)); }
            public boolean addAll(Collection c) { return(col.addAll(c)); }
            public void clear() { col.clear(); }
            public boolean contains(Object o) { return(col.contains(o)); }
            public boolean containsAll(Collection c) { return(col.containsAll(c)); }
            public boolean equals(Object o) { return(col.equals(o)); }
            public int hashCode() { return(col.hashCode()); }
            public boolean isEmpty() { return(col.isEmpty()); }
            public boolean remove(Object o) { return(col.remove(o)); }
            public boolean removeAll(Collection c) { return(col.removeAll(c)); }
            public boolean retainAll(Collection c) { return(col.retainAll(c)); }
            public int size() { return(col.size()); }
            public Object[] toArray() { return(col.toArray()); }
            public Object[] toArray(Object a[]) { return(col.toArray(a)); }
            }


        /**Get user-specified (in properties) level of paralleism.
         */
        public static int getPropLevel()
            {
            // Check if this up-to-date in the cache...
            if(_cache_getPropLevel != 0) { return(_cache_getPropLevel); }
            // ...else force its possibly-expensive recomputation.
            return(_getPropLevel());
            }
        /**Get level of parallelism stipulated in properties, if any.
         * If no level specified, or no usable level, returns -1; else
         * the result is strictly positive.
         * <p>
         * Cached after first access, and synchronized so that the result
         * is computed (and set, to a non-zero value) exactly once.
         */
        private static synchronized int _getPropLevel()
            {
            if(_cache_getPropLevel != 0) { return(_cache_getPropLevel); }
            final String propVal = System.getProperty(SYS_PROP_PAR);
            if((propVal == null) || (propVal.length() == 0))
                {
                if(debug) { System.err.println("[No value for property "+SYS_PROP_PAR+".]"); }
                return(_cache_getPropLevel = -1); // Nothing available...
                }
            try {
                final int v = Integer.parseInt(propVal, 10);
                if(v < 1)
                    {
                    System.err.println("ERROR: Unparsable (non-positive) value for property "+SYS_PROP_PAR+": "+v+".");
                    return(_cache_getPropLevel = -1); // Nothing available...
                    }
                return(_cache_getPropLevel = v);
                }
            catch(NumberFormatException e)
                {
                System.err.println("ERROR: Unparsable non-integer value for property "+SYS_PROP_PAR+": "+propVal+".");
                return(_cache_getPropLevel = -1); // Nothing available...
                }
            }
        /**Private cache for _getPropLevel().
         * Set exactly once, to a non-zero value, under the class lock.
         */
        private static int _cache_getPropLevel; // Initially 0.

        /**If true, the user has explictly turned off parallelism.
         */
        public static boolean noParallelism()
            { return(getPropLevel() == 1); }

        /**Get current estimate of count of system CPUs available.
         * Always strictly positive; units are whole CPUs.
         */
        public static int availableConcurrency()
            {
            final int propLevel = getPropLevel();
            if(propLevel > 0) { return(propLevel); } // Use user-set value.
            return(SYS_PROP_PAR__DEFAULT); // Use default single-CPU assumption.
            }

        /**Current CPU utilisation by tasks under our control, in fractional CPU units.
         * Non-negative, and adjusted under the class lock by
         * _adjustUtilisation().
         * <p>
         * Initially zero.
         */
        private static int _currentUtil;

        /**We allow a little parallel slackness to soack up CPU cycles efficiently.
         * This is in fractional CPU units.
         */
        private static final int _parallelSlackness = Math.max(UTILISATION_MIN,
            (UTILISATION_MAX / 2));

        /**Computes residual fractional CPU units.
         * Never negative.
         */
        private synchronized static int _availableCPU()
            {
            if(noParallelism()) // Behave as if no spare CPU to force at most one thread.
                { return(0); }
            // Multiply up number of CPUs, plus a little parallel slackness,
            // minus work in progress.
            return(Math.max(0, (availableConcurrency() * UTILISATION_MAX) +
                _parallelSlackness - _currentUtil));
            }

        /**Adjusts our notion of global CPU utilisation by tasks we are running.
         * Atomic, uses the class lock, returns the new utilisation
         * in fractional CPU units.
         */
        private synchronized static int _adjustUtilisation(int delta)
            {
            final int result = (_currentUtil += delta);
            if(debug && (delta != 0))
                {
                System.err.println("[ParTasks: total CPU utilisation estimate now: "+
                    ((100 * result) / UTILISATION_MAX)+"%.]");
                }
            //Assert(result >= 0);
            return(result);
            }


        /**CPU utilisation stack per thread.
         * Normally contains a Stack of CPU utilisation values
         * maintained by runTask().
         * <p>
         * When a Thread first takes part in ParTasks, or leaves
         * its outermost ParTasks activity it will have no Stack
         * here, which is taken as indicating a current ParTasks-monitored
         * utilisation of 0.
         * <p>
         * Otherwise, on entry to runTask() the top-most value (taken
         * to be the utilisation of an `outer' task using this thread)
         * is subtracted from the global utilisation, and the new utilisation
         * is pushed on.  On exit from runTask() the top-most value
         * is popped, subtracted from the global utilisation, and the
         * current top value is added to the global utilisation.
         * <p>
         * These stack operations are effectively atomic because only
         * one thread can see the releavant stack.
         */
        private static final ThreadLocal utilStack = new ThreadLocal();

        /**Get current thread's recorded utilisation, in fractional CPU units.
         */
        private static int _getThisThreadUtil()
            {
            final Stack s = (Stack) utilStack.get();
            // Get the old utilisation of this Thread, if any...
            final int oldUtil = ((s == null) || (s.size() < 1)) ?
                0 : (((Integer) s.peek()).intValue());
            //Assert(oldUtil >= 0);
            return(oldUtil);
            }

        /**Exception thrown as a result of a failed Task. */
        public static class ParTasksException extends RuntimeException
            {
            public ParTasksException() { }
            public ParTasksException(String s) { super(s); }
            }

        /**Runs a single task.
         * The predicted CPU utilisation of the task is specified.
         * <p>
         * If the task r throws an exception it will not be propagated,
         * and this routine will throw a ParTasksException.
         * <p>
         * Whatever the current estimated CPU utilisation this will
         * run the task in the current thread.
         * <p>
         * The removes the old utilisation for this thread, if any,
         * from the total load estimate, adds in the new utilisation
         * and pushes it on the ThreadLocal stack of such utilisation.
         * <p>
         * This is effectively a base case.
         * <p>
         * It can also be used by some external spawner of tasks to
         * inform the ParTasks system about other activity in the system.
         */
        public static void runTask(Runnable r, int utilisation)
            throws ParTasksException
            {
            if((utilisation < UTILISATION_ABSMIN) ||
               (utilisation > UTILISATION_ABSMAX))
                { throw new IllegalArgumentException("bad utilisation value"); }

            try {
                Stack s = (Stack) utilStack.get();

                // Get the old utilisation of this Thread, if any...
                final int oldUtil = _getThisThreadUtil();
                //Assert(oldUtil >= 0);

                if(s == null)
                    {
                    // This Thread is not running an outer task...
                    // So create a new Stack ready for our value...
                    s = new Stack();
                    }

                // Ensure that we cannot run out of memory
                // to stack our new utilisation after having fiddled
                // with the global value
                // (ie running out of memory will leave our state intact
                // and just jemmy this task).
                s.ensureCapacity(s.size() + 1);
                utilStack.set(s);

                // Note the new utilisation on the stack...
                s.push(new Integer(utilisation));

                // Adjust global notion of CPU utilisation...
                _adjustUtilisation(utilisation - oldUtil);

                // OK, now actually run the task...
                // This may throw an exception of some sort...
                try { r.run(); }
                catch(Throwable t)
                    {
                    System.err.println("ERROR: unexpected exception: " + t.getMessage());
                    t.printStackTrace();
                    throw new ParTasksException("ERROR: unexpected exception: " + t.getMessage());
                    }
                }
            finally
                {
                // Now re-adjust the global utilisation.

                Stack s = (Stack) utilStack.get();
                // Assert(s.peek() == utilisation);
                s.pop(); // Discard our value...
                // If this Thread is leaving ParTasks, clean up some space.
                if(s.size() == 0) { utilStack.set(s = null); }

                // Get the old utilisation of this Thread, if any...
                final int oldUtil = (s == null) ?
                    0 : (((Integer) s.peek()).intValue());

                // Adjust global notion of CPU utilisation...
                _adjustUtilisation(oldUtil - utilisation);
                }
            }

        /**If true, launch Cloneable tasks in their own threads. */
        private static final boolean _cloneableNewThreads = true;

        /**The interface for a task usable by runCloneableTasks(). */
        private interface CloneableTask extends Runnable, Cloneable
            {
            /**Returns the maximum number of threads sensible for this problem.
             * This should return a strictly-positive number, though any
             * value less than two will indicate that parallelism is
             * probably not useful.
             */
            public int maxThreads();

            /**Allow cloning. */
            public Object clone();
            }

        /**Runs a set of tasks to complete as quickly as possible.
         * The specified task r must be Cloneable and can either
         * be used directly (by calling its run() method),
         * or it can clone()d as many times as required
         * by this routine with each of those tasks then being run()
         * concurrently.  The task need not be clone()able once its
         * run() method has been called.
         * <p>
         * The predicted CPU utilisation of each clone of the task is specified.
         * <p>
         * No value in r[] may be null.
         * <p>
         * If any task (ie r or a clone()) throws an exception
         * then the whole task may not be executed.
         * No thrown exception will be propagated to the caller.
         * <p>
         * Whatever the current estimated CPU utilisation this will
         * run at least one task, in the current thread if need be.
         * <p>
         * Parallelism will only be attempted if:
         * <ul>
         * <li>noParallelism() is false, and,
         * <li>r.maxThreads() is more than 1.
         * </ul>
         */
        private static void runCloneableTasks(final CloneableTask r, final int utilisation)
            throws ParTasksException
            {
            if((utilisation < UTILISATION_ABSMIN) ||
               (utilisation > UTILISATION_ABSMAX))
                { throw new IllegalArgumentException("bad utilisation value"); }

            // See if we can launch some parallelism in new threads...
            if(_cloneableNewThreads &&
               (r.maxThreads() > 1) && !noParallelism())
                {
                // It's at least not ruled out a priori...

                // Make this thread recorded as consuming zero CPU itself,
                // and have it just wait for the new threads to finish.
                final Throwable err[] = new Throwable[1]; // Slot for any errors...
                runTask(new Runnable(){
                    public final void run()
                        {
                        try {
                            // Create and start completely new threads,
                            // and wait for them to complete...
                            final Thread t[] = _launchNewThreads(r, utilisation, err);
                            for(int i = t.length; --i >= 0; )
                                {
                                final Thread th = t[i];
                                while(th.isAlive())
                                    {
                                    try { th.join(); }
                                    catch(InterruptedException e) { }
                                    }
                                }
                            }
                        catch(Throwable t) { err[0] = t; }
                        }
                    },
                    UTILISATION_ABSMIN); // This thread now just waits for the others to finish.
                if(err[0] != null)
                    {
                    err[0].printStackTrace();
                    throw new ParTasksException("Unexpected exception: " + err[0].getMessage());
                    }
                return;
                }

            // Trivial implementation, always using one thread...
            runTask(r, utilisation);
            }

        /**Creates a set of threads to run a set of cloneable tasks.
         * This assumes that there is some parallelism to be had.
         * <p>
         * This is synchronised (ont the class) to ensure a consistent
         * view of available CPU time.
         * <p>
         * This will always launch at least one thread.
         *
         * @param setIfFailure is a 1-element array, whose zeroth element
         *           is set non-null if any of the underlying tasks throws
         *           an exception
         */
        private synchronized static Thread[] _launchNewThreads(final CloneableTask r, final int utilisation,
                                                            final Throwable setIfFailure[])
            {
            if(debug)
                {
                if((utilisation < UTILISATION_ABSMIN) ||
                   (utilisation > UTILISATION_ABSMAX))
                    { throw new IllegalArgumentException("bad utilisation value"); }
                }

            // Launch an appropriate number of threads;
            // always at least one.
            final int nThreads = Math.max(1,
                Math.min(Math.min(r.maxThreads(), SYS_PROP_MAXTHREADS__DEFAULT - tg.activeCount()),
                _availableCPU() / Math.max(utilisation, UTILISATION_MIN)));
if(debug) { System.err.println("[ParTasks: LAUNCHING "+nThreads+" THREADS.]"); } 

if(debug) { System.err.println("[ParTasks: LAUNCHING "+nThreads+" THREADS.]"); }
            final Thread result[] = new Thread[nThreads];
            for(int i = nThreads; --i >= 0; )
                {
                result[i] = new Thread(tg, new Runnable(){
                    public final void run()
                        {
                        // Run as new task (to pin down allocated
                        // CPU utilisation and prevent allocation races)
                        // which will in turn run all the sub-tasks
                        // that this CloneableTask encounters.
                        try { runTask((Runnable) r.clone(), utilisation); }
                        catch(Throwable t) { setIfFailure[0] = t; }
                        }
                    });
                result[i].start();
                }

            return(result);
            }

        /**Possible base class for our internal Cloneable classes.
         */
        private static abstract class TaskBase implements ParTasks.CloneableTask
            {
            /**Returns a clone.
             * There is in fact no (writable) internal state,
             * so we can use Object.clone().
             */
            public Object clone()
                {
                try { return(super.clone()); }
                catch(CloneNotSupportedException e) { throw new Error(); } // Cannot happen...
                }
            }

        /**Functor for getting side-effects only.
         */
        public interface TaskProc
            {
            /**Apply fn to Object for its side-effect. */
            public void fn(Object o);
            }

        /**Functor for filtering objects.
         */
        public interface TaskFilter
            {
            /**If this returns true, the object should appear in the result. */
            public boolean fn(Object o);
            }

        /**Functor for mapping one object to another.
         */
        public interface TaskMapper
            {
            /**Apply fn to Object to get it converted to another. */
            public Object fn(Object o);
            }

        /**Apply a TaskProc to a Collection of objects.
         * The Collection must not be altered while this routine is
         * running.  The TaskProc is applied to all of the objects
         * in the collection.
         */
        public static void apply(Collection cin, final TaskProc f,
                                 final int maxThreads, final int utilisation)
            throws ParTasksException
            {
            if(cin.size() == 0) { return; } // Nothing to do...
            final Collection scin = Collections.synchronizedCollection(cin);
            final Iterator it = scin.iterator();

            final Throwable err[] = new Throwable[1]; // Slot for any errors...

            final class Task extends TaskBase
                {
                /**Returns the maximum number of threads sensible for this problem.
                 * This is the smaller of the number of objects available to work on
                 * and the concurrency limit that we set...
                 */
                public int maxThreads()
                    { return(Math.max(Math.min(maxThreads, scin.size()), 1)); }

                /**Run the task.
                 * Go on getting items to inspect until we run out...
                 */
                public void run()
                    {
                    for( ; ; )
                        {
                        Object o;
                        synchronized(scin) // Protect it.
                            {
                            if(!it.hasNext()) { break; } // Done.
                            o = it.next();
                            }
                        synchronized(o) // Serialise activity on o.
                            {
                            try { f.fn(o); } // Apply functor...
                            catch(Throwable t) { err[0] = t; break; }
                            }
                        }
                    }
                }

            // Now actually run the tasks...
            runCloneableTasks(new Task(), utilisation);

            // Note errors...
            if(err[0] != null)
                {
                err[0].printStackTrace();
                throw new ParTasksException("Unexpected exception: " + err[0].getMessage());
                }
            }

        /**Apply a TaskFilter to a Collection of objects, copying suitable objects to the result Collection.
         * The Collection must not be altered while this routine is
         * running.  The TaskFilter is applied to all of the objects
         * in the collection, and all input items in cin that cause f.fn()
         * to return a boolean are added to result.  The result is wrapped
         * in a synchronizing wrapper...
         * <p>
         * The result might typically be a Set or a Vector.
         */
        public static void apply(Collection cin, final TaskFilter f,
                                 final int maxThreads, final int utilisation,
                                 Collection result)
            {
            if(cin.size() == 0) { return; } // Nothing to do...
            final Collection scin = Collections.synchronizedCollection(cin);
            final Iterator it = scin.iterator();
            final Collection sresult = Collections.synchronizedCollection(result);

            final Throwable err[] = new Throwable[1]; // Slot for any errors...

            final class Task extends TaskBase
                {
                /**Returns the maximum number of threads sensible for this problem.
                 * This is the smaller of the number of objects available to work on
                 * and the concurrency limit that we set...
                 */
                public int maxThreads()
                    { return(Math.max(Math.min(maxThreads, scin.size()), 1)); }

                /**Run the task.
                 * Go on getting items to inspect until we run out...
                 */
                public void run()
                    {
                    for( ; ; )
                        {
                        Object o;
                        synchronized(scin) // Protect it.
                            {
                            if(!it.hasNext()) { break; } // Done.
                            o = it.next();
                            }
                        synchronized(o) // Serialise activity on o.
                            {
                            try { if(f.fn(o)) { sresult.add(o); } } // Apply functor...
                            catch(Throwable t) { err[0] = t; break; }
                            }
                        }
                    }
                }

            // Now actually run the tasks...
            runCloneableTasks(new Task(), utilisation);

            // Note errors...
            if(err[0] != null)
                {
                err[0].printStackTrace();
                throw new ParTasksException("Unexpected exception: " + err[0].getMessage());
                }
            }

        /**Apply a TaskMap to a Set of objects, adding mappings from input to output objects to the result Map.
         * The Collection must not be altered while this routine is
         * running.  The TaskMap is applied to all of the objects
         * in the collection.
         * <p>
         * This applies to a Set to ensure that the input collection has
         * unique keys to use as the key set for the Map result.
         */
        public static void apply(Set cin, final TaskMapper f,
                                 final int maxThreads, final int utilisation,
                                 Map result)
            {
            if(cin.size() == 0) { return; } // Nothing to do...
            final Collection scin = Collections.synchronizedCollection(cin);
            final Iterator it = scin.iterator();
            final Map sresult = Collections.synchronizedMap(result);

            final Throwable err[] = new Throwable[1]; // Slot for any errors...

            final class Task extends TaskBase
                {
                /**Returns the maximum number of threads sensible for this problem.
                 * This is the smaller of the number of objects available to work on
                 * and the concurrency limit that we set...
                 */
                public int maxThreads()
                    { return(Math.max(Math.min(maxThreads, scin.size()), 1)); }

                /**Run the task.
                 * Go on getting items to inspect until we run out...
                 */
                public void run()
                    {
                    for( ; ; )
                        {
                        Object o;
                        synchronized(scin) // Protect it.
                            {
                            if(!it.hasNext()) { break; } // Done.
                            o = it.next();
                            }
                        synchronized(o) // Serialise activity on o.
                            {
                            try { sresult.put(o, f.fn(o)); }
                            catch(Throwable t) { err[0] = t; break; }
                            }
                        }
                    }
                }

            // Now actually run the tasks...
            runCloneableTasks(new Task(), utilisation);

            // Note errors...
            if(err[0] != null)
                {
                err[0].printStackTrace();
                throw new ParTasksException("Unexpected exception: " + err[0].getMessage());
                }
            }
        }
